/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* bcsup.c: support routines for bc program */

#include "eclib.h"


/* error messages; incomprehensible, as required */
char er_const[] = "operand must be a constant, not a polynomial";
char er_arg[] = "arguments must be variables to be expanded";
char er_resrc[] = "@too many allocated structures";
char er_exp[] = "exponent on variable is out of range";
char er_zero[] = "division by zero";
char er_lhs[] = "left side of = must be a variable or var @ exponent";

struct POLY *vb[26] = { /* one letter variables */
&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,
&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,
&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,&pzero,&pzero};
char vbr[26] = { /* is a root of the polynomial implied */
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
struct POLY *dotval = &pzero; /* what was last displayed */

char err_seen; /* semantic or syntactic error seen */
char intdiv; /* integer divide, or leave rational */
char is_square; /* is the argument passed to isqr() a perfect square */

/* fixed integers */
struct VINT izero = { 0, 0, {0} }; /* 0 */
struct VINT ione = { 1, 0, {1} }; /* 0 */
struct VINT digit = { 1, 0, {1} }; /* single digit number */
struct VINT ibase = { 1, 0, {10} }; /* input redix value */
struct VINT obase = { 1, 0, {10} }; /* output redix value */
struct VINT *modulus = &izero; /* modulus for modular arithmetic */

/* fixed polynomials */

struct POLY pzero = { /* the 0 polynomial */
0, /* no link */  1, /* one term */
{ {
&izero, /* coefficient */
0, /* number of variables */
0, /* degree */
{0}, /* variables */
{0}, /* exponents */
}}};

struct POLY pone = { /* the 1 polynomial */
0, /* no link */  1, /* one term */
{ {
&ione, /* coefficient */
0, /* number of variables */
0, /* degree */
{0}, /* variables */
{0}, /* exponents */
}}};

struct POLY pconst = { /* a constant polynomial */
0, /* no link */  1, /* one term */
{ {
0, /* coefficient */
0, /* number of variables */
0, /* degree */
{0}, /* variables */
{0}, /* exponents */
}}};

struct POLY pvar = { /* single variable polynomial */
0, /* no link */  1, /* one term */
{ {
&ione, /* coefficient */
1, /* number of variables */
0, /* degree */
{0}, /* variables */
{0}, /* exponents */
}}};

/* end of global variables */

#define MAXALLOC 2000

/* static variables */
static char *reslist[MAXALLOC]; /* list of allocated resources */
static int nalloc = 0;  /* number of allocated resources */


/* allocate/free routines */

/* get new integer */
struct VINT *newnum(n) /*###*/
{
struct VINT *s;

if(nalloc == MAXALLOC) eprint(er_resrc);
s = (struct VINT *)emalloc(n+n+4);
reslist[nalloc++] = (char*)s;
s->size = n;
return s;
}

/* get new polynomial */
struct POLY *newpoly(n) /*###*/
{
struct POLY *s;

if(nalloc == MAXALLOC) eprint(er_resrc);
s = (struct POLY *)emalloc(sizeof(struct POLY) + sizeof(struct TERM)*(n-1));
reslist[nalloc++] = (char*)s;
/* set it up as though it is a constant */
s->next = 0;
s->nterms = n;
s->terms[0].coef = 0;
s->terms[0].nvars = 0;
s->terms[0].degree = 0;
return s;
}

/* is integer tied to a polynomial? */
tiedto(p, s) /*###*/
struct POLY *p;
struct VINT *s;
{
register struct TERM *t;

for(;p; p=p->next)
{
if((struct VINT *)p == s) return 1;
for(t=p->terms; t<p->terms + p->nterms; ++t)
if(s == t->coef) return 1;
}

return 0;
}

/* free bc entity on the resource list */
bcfree(t) /*###*/
struct VINT *t;
{
register i;

/* not terribly efficient, but we assume malloc/free time */
/* is small compared with computation time. */
if(t == modulus) return;
if(tiedto(dotval, t)) return;
for(i=0; i<26; ++i)
if(tiedto(vb[i], t)) return;

for(i=0; i<nalloc; ++i)
if(reslist[i] == (char*)t) { reslist[i] = reslist[--nalloc]; free(t); return; }
eprint("@cannot free resource");
} /* bcfree */

/* free a known intermediate polynomial */
polyfree(p) /*###*/
struct POLY *p;
{
register struct TERM *t;

if(p == &pzero || p == &pone) return;
for(t=p->terms; t<p->terms + p->nterms; ++t)
bcfree(t->coef);
bcfree(p);
} /* polyfree */

/* clean up, free all loose resources */
allfree() /*###*/
{
register i;
int last;

last = nalloc;
for(i=0; i<nalloc; ++i)
{
bcfree(reslist[i]);
if(last == nalloc) continue; /* not freed */
--i; /* new one in this position, free it */
last = nalloc;
}
}


/* integer arithmetic */

/* divide by 2, t >>= 1 */
void ishr(t) /*###*/
struct VINT *t;
{
register i;
unsigned size;

size = t->size & 0x7fff;
if(!size) return;

for(i=0; i<size-1; ++i)
{
t->value[i] >>= 1;
if(t->value[i+1] & 1) t->value[i] |= 0x8000;
}
t->value[i] >>= 1;

if(t->value[i]) return;
if(--t->size == 0x8000) t->size = 0;
}

/* multiply by 2, t <<= 1 */
void ishl(t) /*###*/
struct VINT *t;
{
register i;
unsigned size;

size = t->size & 0x7fff;
if(!size) return;

i = size-1;
if(t->value[i] & 0x8000) { t->value[i+1] = 1; ++t->size; }
for(; i>0; --i)
{
t->value[i] <<= 1;
if(t->value[i-1] & 0x8000) t->value[i] |= 1;
}
t->value[0] <<= 1;
}

/* reduce argument by the modulus */
/* bcfrees the argument if necessary */
struct VINT *modcheck(t) /*###*/
struct VINT *t;
{
register n;
struct VINT *quot, *rem;

if(!modulus->size) return t;
if(!t->size) return t;
n = t->size & 0x7fff;
quot = newnum(n); /* allocating more than we need */
n += 2; /* 2 extra words for long division routine */
if(modulus->size > n) n = modulus->size;
rem = newnum(n);
vdiv(t, modulus, quot, rem);
bcfree(t);
bcfree(quot);
return rem;
}

/* integer add */
struct VINT *iadd(a, b) /*###*/
struct VINT *a, *b;
{
register struct VINT *t;
int n;

n = (a->size & 0x7fff);
if((b->size & 0x7fff) > n) n = (b->size & 0x7fff);
t = newnum(n+1);
vadd(a, b, t);
return modcheck(t);
}

/* integer subtract */
struct VINT *isub(a, b) /*###*/
struct VINT *a, *b;
{
register struct VINT *t;
int n;

n = (a->size & 0x7fff);
if((b->size & 0x7fff) > n) n = (b->size & 0x7fff);
t = newnum(n+1);
vsub(a, b, t);
return modcheck(t);
}

/* integer multiply */
struct VINT *imul(a, b) /*###*/
struct VINT *a, *b;
{
register struct VINT *t;
int n;

n = ((a->size + b->size) & 0x7fff);
t = newnum(n);
vmul(a, b, t);
return modcheck(t);
}

/* integer remainder */
struct VINT *irem(a, b) /*###*/
struct VINT *a, *b;
{
register struct VINT *quot,  *rem;
int n;

if(!b->size) { eprint(er_zero); return &izero; }
n = a->size & 0x7fff;
quot = newnum(n); /* allocating more than we need */
n += 2; /* 2 extra words for long division routine */
if((b->size & 0x7fff) > n) n = (b->size & 0x7fff);
rem = newnum(n);
vdiv(a, b, quot, rem);
bcfree(quot);
return rem;
}

/* integer divide */
struct VINT *idiv(a, b) /*###*/
struct VINT *a, *b;
{
register struct VINT *quot,  *rem;
int n;

if(modulus->size) /* divide with modulus */
return imul(a, modinv(b, modulus));

if(!b->size) { eprint(er_zero); return &izero; }
n = a->size & 0x7fff;
quot = newnum(n); /* allocating more than we need */
n += 2; /* 2 extra words for long division routine */
if((b->size & 0x7fff) > n) n = (b->size & 0x7fff);
rem = newnum(n);
vdiv(a, b, quot, rem);
bcfree(rem);
return quot;
}

/* take gcd of the arguments */
struct VINT *igcd(a, b) /*###*/
struct VINT *a, *b;
{
unsigned n, n1; /* size */
struct VINT *s,*t, *u, *c; /* gcd values */
struct VINT *swap; /* for swaps */

/* make copies, so we need not free arguments */
n = a->size & 0x7fff;
n1 = b->size & 0x7fff;
if(n1 > n) n = n1;
s = newnum(n+2);
t = newnum(n+2);
u = newnum(n+2);
c = newnum(n);
vadd(a, &izero, s);
vadd(b, &izero, t);

while(t->size)
{
vdiv(s, t, c, u);
swap = s, s = t, t = u, u = swap;
}

bcfree(t), bcfree(u), bcfree(c);
return s;
}

/* compute the inverse of x mod m.
 * this is equivalent to solving the equation x*s + m*t = 1,
 * where s and t are integers, and s is the inverse of x.
 * efficient procedures exist for solving such equations, exploiting 
 * the power of the gcd algorithm.  While computing gcd(x,m)
 * (which must be 1), the quotients can be used to
 * produce the inverse of x.
 */
struct VINT *modinv(x, m) /*###*/
struct VINT *x, *m;
/* m is the modulus, which we assume is positive */
/* we also assume the routine modcheck() reduces mod m */
{
unsigned n; /* size */
struct VINT *s,*t, *u; /* gcd values */
struct VINT *a, *b; /* coeficients */
struct VINT *q, *r, *o, *c; /* temporaries */
struct VINT *swap; /* for swaps */

/* make copies, so we need not free arguments */
n = m->size;
s = newnum(n+2);
t = newnum(n+2);
u = newnum(n+2);
vadd(modcheck(x), &izero, t);
vadd(m, &izero, s);
a = newnum(n);
a->size = 0;
b = newnum(n);
b->size = 1;
b->value[0] = 1;
o = newnum(n+n+1);
c = newnum(n);
q = newnum(n+n);
r = newnum(n+n+2);

while(t->size)
{
vdiv(s, t, c, u);
swap = s, s = t, t = u, u = swap;
vmul(c, b, q);
vsub(a,q,o);
vdiv(o, m, c, r);
swap = a, a = b, b = swap;
vadd(r, &izero, b);
}

bcfree(o), bcfree(q), bcfree(r), bcfree(c),bcfree(b),bcfree(t), bcfree(u);
if(s->size != 1 || s->value[0] != 1)
{
bcfree(a);
eprint("argument and modulus are not relatively prime");
idisp(s);
printf("\n");
bcfree(s);
return &izero;
}

bcfree(s);
return a;
} /* modinv */

/* square root, newton's method */
struct VINT *isqr(x) /*###*/
struct VINT *x;
{
struct VINT *a, *b, *t;
register i;
int n;
short save_bit;

is_square = 1;
if(!x->size) return &izero;

save_bit = x->size & 0x8000;
x->size &= 0x7fff; /* temporarily make it positive */
n = x->size / 2 + 1;
a = newnum(n);
b = newnum(n);
t = newnum(x->size+2);
a->size = n;
a->value[n-1] = 1;

while(1)
{
is_square = 1; /* assumed ok */
vdiv(x, a, b, t);
if(t->size) is_square = 0;

/* test for equal operands */
vsub(b, a, t);
if(!t->size) break;
is_square = 0;
if((t->size&0x7fff) == 1 && t->value[0] == 1) break;

vadd(a, b, t);
ishr(t);
vadd(t, &izero, a);
}

n = t->size;
bcfree(t);
if(n < 0) t = a, a = b, b = t;
bcfree(b);
x->size |= save_bit;
return a;
}

/* factorial */
struct VINT *vbang(x) /*###*/
struct VINT *x;
{
struct VINT *s, *t;
int n = 0;

if(x->size) n = x -> value[0];
s = newnum(1);
s->value[0]=1;

for(; n; --n)
{
digit.value[0] = n;
t = imul(s, &digit);
bcfree(s);
s = t;
}
return s;
}

/* raise number to an integer power */
struct VINT *vexp(base, t) /*###*/
struct VINT *base, *t;
{
register i;
struct VINT *b, *s;

if(t->size < 0) return &izero; /* negative exponent */
if(!base->size) return &izero;
/* make copies */
s = newnum(t->size);
vadd(t, &izero, s);
t = s;
base = iadd(base, &izero);
b = newnum(1);
b -> value[0] = 1;

while(t->size)
{
if(t->value[0] & 1)
{
s = imul(b, base);
bcfree(b);
b = s;
}
s = imul(base, base);
bcfree(base);
base = s;
ishr(t);
}

bcfree(t);
bcfree(base);
return b;
}

/* binary coefficient */
struct VINT *vchoose(x, k) /*###*/
struct VINT *x, *k;
{
unsigned short i;
struct VINT x1;
struct VINT *s, *r, *t;

if((x->size | k->size) < 0) return &izero;
if(!x->size) return &izero;
if(x->size > 1 || k->size > 1)
{
eprint("binary coefficient too large");
return &izero;
}
i = 0;
if(k->size) i = k->value[0];
x1 = *x;
if(i > x1.value[0]) return &izero;
s = newnum(1);
s->value[0] = 1;
if(i+i > x1.value[0]) i = x1.value[0] - i; /* symmetry */
digit.value[0] = 1;

while(i)
{
t = imul(s, &x1);
bcfree(s);
s = idiv(t, &digit);
bcfree(t);
--x1.value[0];
++digit.value[0];
--i;
}
return s;
}

/* print passed integer */
idisp(t)
struct VINT *t;
{
register i, n;
int column;
char *s;
struct VINT *t1, *t2;

n = t->size;
if(n < 0) printf("-");
n &= 0x7fff;
/* make copy, so we can remove the sign bit with impunity */
t1 = iadd(t, &izero);
t = t1;
n = t->size;
n &= 0x7fff;
t->size = n;

if(obase.value[0] >= 10)
s = emalloc(5*n + 1); /* more room than the number actually needs */
else
s = emalloc(16*n + 1); /* base 2,worst case */

for(i=0; n; ++i) {
t1 = newnum(n);
t2 = newnum(n+2);
vdiv(t, &obase, t1, t2);
s[i] = '0';
if(t2->size) {
s[i] = t2->value[0] + '0';
if(s[i] > '9') s[i] += 'A'-'9'-1;
}
bcfree(t2);
bcfree(t);
t = t1;
n = t->size;
}

if(!i) printf("0");
for(column=0; i--; ++column) {
if(column == 50)
column=0, printf("\\\n");
putchar(s[i]);
}
free(s);
} /* idisp */

struct VINT *string_to_vint(s_num)
char *s_num;
{
char c;
struct VINT *s, *t = &izero;

/* convert string to VINT, not terribly efficient */
while(c = *s_num++) {
if(isalpha(c)) {
c = toupper(c);
c -= 'A'-'9'-1;
}
digit.value[0] = (c - '0');
s = newnum(t->size + 1);
vmul(t, &ibase, s);
bcfree(t);
t = newnum(s->size + 1);
if(digit.value[0])
vadd(s, &digit, t);
else
vadd(s, &izero,t);
bcfree(s);
} /* end string */

return t;
} /* string_to_vint */

struct VINT *long_to_vint(l)
long l;
{
struct VINT *t = &izero;

if(l) {
t = newnum(2);
t->value[0] = l;
t->size = 1;
if(t->value[1] = (l>>16))
t->size = 2;
}
return t;
} /* long_to_vint */

