/***********************************************************************
|  Basic calculator with polynomials
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* bcfactor.c: primality and factorization */

#include "bcgram.h"

#define STRENGTH 8 /* number of iterations of the strong pseudoprime test */
#define MAXPOWER 18 /* iterate 2^MAXPOWER times before giving up factoring */


/* a few simple primes */
static unsigned short prms[54] = {
2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 
41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 
113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 
181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 251, 
};

/* strong pseudoprime test */
/* assumes argument > 1 */
isprime(p) /*###*/
struct VINT *p;
{
struct VINT *s, *r;
struct VINT *b, *c, *t, *q, *qr;
int n;
register i;
int npowers, np;
int trial;

n = p->size;
if(n == 1 && p->value[0] == 2) return 1;
if(!(p->value[0] & 1)) return 0; /* even */

/* try dividing by small primes */
s = newnum(n);
r = newnum(n+2);
for(i=1; i<54; ++i)
{
digit.value[0] = prms[i];
vdiv(p, &digit, s, r);
if(!r->size) /* goes in evenly? */
{
bcfree(r);
n = 0;
if(s->size == 1 && s->value[0] == 1) n = 1;
bcfree(s);
return n;
}
} /* end loop over small primes */
if(n == 1) /* less than 256^2 */
{
bcfree(r);
bcfree(s);
return 1; /* no prime divisors */
}

/* pseudo prime test here */
/* remove the factors of 2 from (p-1) */
vsub(p, &ione, s);
npowers = 0;
while(!(s->value[0] & 1)) { ishr(s); ++npowers; }

t = newnum(n);
b = newnum(n);
c = newnum(n);
q = newnum(n+n);
qr = newnum(n+n+2);

/* Try several prime values, starting with 13.
 * Each time through, we should really select a value at random,
 * but I am lazy, so I just use a small prime.
 * This may weaken the strong pseudoprime test; I just don't know.
 */

for(trial=5; trial<STRENGTH+5; ++trial)
{
c->size = 1, c->value[0] = prms[trial];

/* c^s mod p */
vadd(s, &izero, t);
b->size = 1, b->value[0] = 1;
while(t->size)
{
if(t->value[0] & 1)
{
vmul(b, c, q);
vdiv(q, p, r, qr);
vadd(qr, &izero, b);
}
vmul(c, c, q);
vdiv(q, p, r, qr);
vadd(qr, &izero, c);

ishr(t);
} /* end loop for t nonzero */

if(b->size == 1 && b->value[0] == 1)
goto nexttrial; /* looks ok */
vsub(p, b, r);
if(r->size == 1 && r->value[0] == 1)
goto nexttrial; /* looks ok */

np = npowers;
while(--np > 0)
{
/* square the value */
vmul(b, b, q);
vdiv(q, p, r, qr);
vadd(qr, &izero, b);

if(b->size == 1 && b->value[0] == 1)
{
/* not prime, free things and return */
bcfree(r), bcfree(s), bcfree(t), bcfree(b), bcfree(c), bcfree(q), bcfree(qr);
return 0;
}
vsub(p, b, r);
if(r->size == 1 && r->value[0] == 1)
goto nexttrial; /* looks ok */
}

/* gone all the way of to a^((p-1)/2) without a +-1 */
/* either fails fermat's little theorem or has too many square roots of 1 */
bcfree(r), bcfree(s), bcfree(t), bcfree(b), bcfree(c), bcfree(q), bcfree(qr);
return 0;
nexttrial: ;
} /* end of trials */

/* all tests pass */
bcfree(r), bcfree(s), bcfree(t), bcfree(b), bcfree(c), bcfree(q), bcfree(qr);
return 1; /* p is probably prime */
}


/* factor a number.   Distroys the argument.
 * Pollard Rho method.
 * Not meant for serious factoring, just something easy to program up.
 * assumes argument is odd and > 1.
 */
factor(t) /*###*/
struct VINT *t;
{
/* set the constant term for our quadratic polynomial */
unsigned short base = 29879;
struct VINT *quot, *rem;
struct VINT *a, *b, *swap;
struct VINT *x, *x0, *square;
short size;
long count, power2;
struct VINT *bulk; /* for multiplying several differences together */
	/* and running gcd on the lot of them simultaneously */
struct VINT *old_x;
char bulkflag;
char primeflag;

/* allocate everything once */
size = t->size;
quot = newnum(size);
rem = newnum(size+2);
x0 = newnum(size);
x = newnum(size+size+2);
square = newnum(size+size+2);
bulk = newnum(size+size+2);
old_x = newnum(size+size+2);
a = newnum(size+2);
b = newnum(size+2);

/* Try the small divisors */
primeflag = 1;
for(digit.value[0] = 3; digit.value[0] < 65535; digit.value[0] += 2)
{
if(primeflag && isprime(t)) break;
primeflag = 0;

vdiv(t, &digit, quot, rem);
if(rem->size) continue;
idisp(&digit);
printf("*");
vadd(quot, &izero, t);
primeflag = 1;
digit.value[0] -= 2; /* try it again */
}

if(primeflag) { idisp(t); printf("\n"); return; }

while(1) /* keep splitting t */
{
x->size = 1, x->value[0] = 5;  /* seed the sequence */
power2 = 1;
bulkflag = 0;
bulk->size = 1, bulk->value[0] = 1;

for(count = 1; count < (1L<<MAXPOWER); ++count)
{
if(count == power2)
{
power2 <<= 1;
vadd(x, &izero, x0);
if(power2 == 0x400L) bulkflag = 1;
}
if(bulkflag && (char)count == (char)0)
vadd(x, &izero, old_x);

vmul(x, x, square);
digit.value[0] = base;
vadd(square, &digit, x);
swap = x, x = square, square = swap;
vdiv(square, t, quot, x);

/* run gcd(x-x0, t) */
vsub(x, x0, b);
if(!b->size) { base += 2; break; }
if(b->size < 0)
{
vadd(b, t, a);
swap = a, a = b, b = swap;
}

/* if bulkflag is set, just build them up until you have 256 of them */
if(bulkflag)
{
vmul(b, bulk, square);
vdiv(square, t, quot, bulk);
if((char)count != (char)0xff) continue;
if((count & 0xfff) == 0)
time((long *)0); /* allow the user to break */
}

vadd(t, &izero, a);
if(bulkflag) vadd(bulk, &izero, b);

while(b->size)
{
if(b->size == 1 && b->value[0] == 1) goto next_cnt;
vdiv(a, b, quot, rem);
swap = a, a = b, b = rem, rem = swap;
}

/* a real gcd */
vdiv(t, a, b, rem);

if(!isprime(a))
{

/* I know, I'm throwing away information here, but it hardly ever happens,
 * and when it does, we have already spent so much time finding the composite
 * factor, that spending a litttle more to find one of its primes won't
 * be noticed, and I am too lazy to set up storage buffers and whatnot.
 */
if(!bulkflag) { base += 2; break; }

/* backtrack */
bulkflag = 0;
vadd(old_x, &izero, x);
count -= 256;
goto next_cnt;
}

idisp(a);
printf("*");

if(isprime(b)) { idisp(b); printf("\n"); return; }

vadd(b, &izero, t);
break; /* find the next prime factor */

next_cnt: ;
} /* end of trials */

if(count < (1L<<MAXPOWER)) continue;
printf("???\n");
return;
} /* loop splitting the composite number */

/* not reached */
}

